/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 * File: session.c
 *
 * Created: Wed May 14 17:22:09 2003
 *
 * $Id: session.c,v 1.1 2003/12/13 20:59:15 mhassler Exp $
 */

#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "defines.h"
#include "session.h"
#include "dns.h"
#include "body.h"
#include "buf.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif /* DMALLOC */

extern int filter_proxy;
extern int filter_cookies;
extern int filter_java;
extern int filter_activex;
extern int filter_popups;

struct session *session_list = 0;

int id = 0;

int session_count()
{
	int res = 0;
	struct session *ptr = session_list;

	while(ptr)
	{
		res++;
		ptr = ptr->next;
	}
	return res;
}

void push_state(struct connection *con, int state)
{
	struct stack *ptr = malloc(sizeof(struct stack));
	if(!ptr)
		return;

	ptr->state = state;
	NODE_ADD(con->bodystate, ptr);
}

int pop_state(struct connection *con)
{
	struct stack *ptr = con->bodystate;
	int state;

	if(!ptr)
		return -1;

	state = ptr->state;
	NODE_DEL(con->bodystate, ptr);
	free(ptr);
	return state;
}

int peek_state(struct connection *con)
{
	struct stack *ptr = con->bodystate;

	if(!ptr)
		return -1;

	return ptr->state;
}

int session_accept(int lsok)
{
	struct sockaddr addr;
	socklen_t addrlen;
	struct session *ptr;
	int sok;
	addrlen = sizeof(struct sockaddr);

	sok = accept(lsok, &addr, &addrlen);
	if(sok < 0)
		return -1;

	ptr = malloc(sizeof(struct session));
	if(!ptr)
		return -1;
	memset(ptr, 0, sizeof(struct session));

	ptr->client.in.data = malloc(BUF_SIZE);
	if(!ptr->client.in.data)
	{
		free(ptr);
		return -1;
	}

	ptr->client.out.data = malloc(BUF_SIZE);
	if(!ptr->client.out.data)
	{
		free(ptr);
		free(ptr->client.in.data);
		return -1;
	}

	ptr->server.out.data = malloc(BUF_SIZE);
	if(!ptr->server.out.data)
	{
		free(ptr);
		free(ptr->client.in.data);
		free(ptr->client.out.data);
		return -1;
	}

	ptr->client.sok = sok;
	ptr->server.sok = -1;
	ptr->dns_sok = -1;
	ptr->id = id++;

	NODE_ENQUEUE(session_list, ptr);

	dprintf("Connection accepted\n");

	return sok;
}

int session_fill_rfds(fd_set *rfds, int *maxfd)
{
	struct session *ptr = session_list;

	while(ptr)
	{
		if(ptr->client.sok >= 0 && ptr->client.in.data && ptr->client.in.size < BUF_SIZE)
		{
			FD_SET(ptr->client.sok, rfds);
			if(ptr->client.sok > *maxfd)
				*maxfd = ptr->client.sok;
		}
		if(ptr->server.sok >= 0 && ptr->server.in.data && ptr->server.in.size < BUF_SIZE)
		{
			FD_SET(ptr->server.sok, rfds);
			if(ptr->server.sok > *maxfd)
				*maxfd = ptr->server.sok;
		}
		if(ptr->dns_sok >= 0)
		{
			FD_SET(ptr->dns_sok, rfds);
			if(ptr->dns_sok > *maxfd)
				*maxfd = ptr->dns_sok;
		}
		ptr = ptr->next;
	}
	return 0;
}

int session_fill_wfds(fd_set *wfds, int *maxfd)
{
	struct session *ptr = session_list;

	while(ptr)
	{
		if(ptr->client.sok >= 0 && ptr->client.out.data && ptr->client.out.size > 0)
		{
			FD_SET(ptr->client.sok, wfds);
			if(ptr->client.sok > *maxfd)
				*maxfd = ptr->client.sok;
		}
		if(ptr->server.sok >= 0 && ptr->server.out.data && ptr->server.out.size > 0)
		{
			FD_SET(ptr->server.sok, wfds);
			if(ptr->server.sok > *maxfd)
				*maxfd = ptr->server.sok;
		}
		ptr = ptr->next;
	}
	return 0;
}

int session_close(struct session *ses)
{
	struct connection *client = &ses->client;
	struct connection *server = &ses->server;

	dprintf("Closing the client connection of ses %d (0x%X)!\n", ses->id, ses);

	close(client->sok);
	if(client->in.data)
	{
		free(client->in.data);
		client->in.data = 0;
	}
	if(client->out.data)
	{
		free(client->out.data);
		client->out.data = 0;
	}
	memset(client, 0, sizeof(struct connection));
	client->sok = -1;
	while(client->bodystate)
		pop_state(client);

	if(server->in.data)
	{
		free(server->in.data);
		server->in.data = 0;
	}
	server->in.size = 0;

	session_close_server(ses);

	return 0;
}

int session_close_server(struct session *ses)
{
	struct connection *client = &ses->client;
	struct connection *server = &ses->server;

	dprintf("Closing the server connection of ses %d!\n", ses->id);

	close(server->sok);
/*	if(server->in.data)
	{
		free(server->in.data);
		server->in.data = 0;
	}*/
	if(server->out.data)
	{
		free(server->out.data);
		server->out.data = 0;
	}
	server->out.size = 0;
//	memset(server, 0, sizeof(struct connection));
	server->sok = -1;
	while(server->bodystate)
		pop_state(server);

	client->in.size = 0;
	if(client->in.data)
	{
		free(client->in.data);
		client->in.data = 0;
	}

	return 0;
}

int session_connect_server(struct connection *server)
{
	int sok;
	struct sockaddr_in sin;

	sok = socket(AF_INET, SOCK_STREAM, 0);
	if(sok < 0)
		return -1;

	sin.sin_family = AF_INET;
	sin.sin_port = htons(server->port);
	sin.sin_addr.s_addr = htonl(server->ip);
	if(connect(sok, (struct sockaddr*)&sin, sizeof(struct sockaddr_in)) < 0)
		return -1;

	if(!server->in.data)
	{
		server->in.data = malloc(BUF_SIZE);
		if(!server->in.data)
			return -1;
	}

	if(!server->out.data)
	{
		server->out.data = malloc(BUF_SIZE);
		if(!server->out.data)
		{
			free(server->out.data);
			return -1;
		}
	}

	server->sok = sok;

	return sok;
}

int session_iterate_read(fd_set *fds)
{
	struct session *ptr = session_list;

	while(ptr)
	{
		if(ptr->client.sok >= 0 && ptr->client.in.data && ptr->client.in.size < BUF_SIZE)
			if(FD_ISSET(ptr->client.sok, fds))
				session_read_client(ptr);
		if(ptr->server.sok >= 0 && ptr->server.in.data && ptr->server.in.size < BUF_SIZE)
			if(FD_ISSET(ptr->server.sok, fds))
				session_read_server(ptr);
		if(ptr->dns_sok >= 0)
			if(FD_ISSET(ptr->dns_sok, fds))
			{
				unsigned long ip;
				ptr->server.ip = dns_read(ptr->dns_sok);
				close(ptr->dns_sok);
				ptr->dns_sok = -1;
				ip = htonl(ptr->server.ip);
				dprintf("Trying to connect to '%s:%d'\n", inet_ntoa(*(struct in_addr*)&ip), ptr->server.port);
				session_connect_server(&ptr->server);
			}
		ptr = ptr->next;
	}
	return 0;
}

int session_iterate_write(fd_set *fds)
{
	struct session *ptr = session_list;

	while(ptr)
	{
		if(ptr->client.sok >= 0 && ptr->client.out.size)
			if(FD_ISSET(ptr->client.sok, fds))
				session_write_connection(&ptr->client);
		if(ptr->server.sok >= 0 && ptr->server.out.size)
			if(FD_ISSET(ptr->server.sok, fds))
				session_write_connection(&ptr->server);
		ptr = ptr->next;
	}
	return 0;
}

int session_read_client(struct session *ses)
{
	struct connection *client = &ses->client;
	int res;

	dprintf("Something on the client socket of ses %d...\n", ses->id);

	if(client->in.size >= BUF_SIZE)
		return 0;	// buffer is full

	res = read(client->sok, &client->in.data[client->in.size], BUF_SIZE - client->in.size);
	if(res < 0)
	{
		dprintf("Read error %d\n", errno);
		session_close(ses);
	}
	else if(!res)
		session_close(ses);
	else
	{
		dprintbuf(&client->in.data[client->in.size], res);
		client->in.size += res;
	}
	return 0;
}

int session_write_connection(struct connection *conn)
{
	int res;

	dprintf("Trying to write to connection...\n");

	res = write(conn->sok, conn->out.data, conn->out.size);
	dprintf("Written:\n");
	dprintbuf(conn->out.data, res);
	dprintf("\n---\n");
	memmove(conn->out.data, &conn->out.data[res], conn->out.size-res);
	conn->out.size -= res;

	return 0;
}

int session_read_server(struct session *ses)
{
	struct connection *server = &ses->server;
	int res;

	dprintf("Something on the server socket of ses %d...\n", ses->id);

	if(server->in.size >= BUF_SIZE)
		return 0;	// buffer is full

	res = read(server->sok, &server->in.data[server->in.size], BUF_SIZE - server->in.size);
	if(res < 0)
	{
		dprintf("Read error %d\n", errno);
		session_close_server(ses);
	}
	else if(!res)
		session_close_server(ses);
	else
	{
		dprintbuf(&server->in.data[server->in.size], res);
		server->in.size += res;
	}

	return 0;
}

int session_iterate_parse_client()
{
	struct session *ptr = session_list;

	while(ptr)
	{
		if(ptr->client.sok >= 0)
			client_parse(ptr);
		ptr = ptr->next;
	}
	return 0;
}

int session_iterate_parse_server()
{
	struct session *ptr = session_list;

	while(ptr)
	{
		if(ptr->server.in.size)
			server_parse(ptr);
		ptr = ptr->next;
	}
	return 0;
}

int session_resp_close_client(struct connection *client, char *response)
{
	buf_printf(&client->out, "HTTP/1.1 %s\r\n", response);
	buf_printf(&client->out, "Content-Type: text/html\r\n");
	buf_printf(&client->out, "Connection: close\r\n\r\n");
	buf_printf(&client->out, "<HTML>\r\n<HEAD><TITLE>%s</TITLE></HEAD>\r\n", response);
	buf_printf(&client->out, "<BODY>\r\n<H2>%s</H2>\r\n", response);
	buf_printf(&client->out, "<HR>\r\n</BODY>\r\n</HTML>\r\n");

	client->in.size = 0;
	if(client->in.data)
	{
		free(client->in.data);
		client->in.data = 0;
	}
	return 0;
}

int session_resolve_host(struct session *ses, char *buf, int len)
{
	char host[256];
	struct in_addr addr;

	strncpy(host, buf, min(255, len));
	host[min(255, len)] = 0;
	if(!inet_aton(host, &addr))
		ses->dns_sok = dns_query(buf, len);
	else
	{
		ses->server.ip = ntohl(addr.s_addr);
		dprintf("Trying to connect to '%s:%d'\n", inet_ntoa(addr), ses->server.port);
		session_connect_server(&ses->server);
	}
	return 0;
}

int session_extract_host(struct session *ses, char *host, int len)
{
	char *col_pos = 0;
	char port[7];

	if(host[len-1] == 10)
		len--;
	if(host[len-1] == 13)
		len--;

	col_pos = memchr(host, ':', len);
	if(col_pos)
	{
		// HOST:PORT
		dprintf("Resolve host '");
		dprintbuf(host, col_pos-host);
		dprintf("', port '");
		dprintbuf(col_pos+1, len-(col_pos-host)-1);
		dprintf("'\n");
		strncpy(port, col_pos+1, min(6, len-(col_pos-host)-1));
		port[min(6, len-(col_pos-host)-1)] = 0;
		ses->server.port = atoi(port);

		session_resolve_host(ses, host, col_pos-host);
	}
	else
	{
		// HOST only
		dprintf("Resolve host '");
		dprintbuf(host, len);
		dprintf("'\n");
		if(!ses->server.port)
			ses->server.port = 80;

		session_resolve_host(ses, host, len);
	}
	return 0;
}

int session_handle_method(struct session *ses, char *start, int len)
{
	struct connection *client = &ses->client;
	char *col_pos, *sp_pos, port[7];

	dprintf("Determining the HTTP method\n");

	if(!strncasecmp(start, "OPTIONS", min(7, len)))
		client->method = MET_OPTIONS;
	else if(!strncasecmp(start, "GET", min(3, len)))
		client->method = MET_GET;
	else if(!strncasecmp(start, "HEAD", min(4, len)))
		client->method = MET_HEAD;
	else if(!strncasecmp(start, "POST", min(4, len)))
		client->method = MET_POST;
	else if(!strncasecmp(start, "PUT", min(3, len)))
		client->method = MET_PUT;
	else if(!strncasecmp(start, "DELETE", min(6, len)))
		client->method = MET_DELETE;
	else if(!strncasecmp(start, "TRACE", min(5, len)))
		client->method = MET_TRACE;
	else if(!strncasecmp(start, "CONNECT", min(7, len)))
	{
		client->method = MET_CONNECT;
		col_pos = memchr(start+7, ':', len-7);
		if(col_pos)
		{
			sp_pos = memchr(col_pos+1, ' ', len-(col_pos-start)-1);
			if(sp_pos)
			{
				strncpy(port, col_pos+1, sp_pos-col_pos-1);
				port[sp_pos-col_pos-1] = 0;
				ses->server.port = atoi(port);
			}
		}
	}
	else
		client->method = MET_ERROR;

	dprintf("Method is %i\n", client->method);
	return 0;
}

int client_parse_header(struct session *ses, char *hdr, int len)
{
	int copy = 1;

	if(!strncasecmp(hdr, "Host: ", min(6, len)))
		session_extract_host(ses, hdr+6, len-6);
	else if(!strncasecmp(hdr, "Cookie: ", min(8, len)) && filter_cookies)
		copy = 0;
	else if(!strncasecmp(hdr, "Accept-Encoding: ", min(17, len)))
		copy = 0;
	else if(!strncasecmp(hdr, "Keep-Alive: ", min(12, len)))
		copy = 0;
	else if(!strncasecmp(hdr, "Connection: ", min(12, len)))
		copy = 0;

	if(copy)
	{
		if(ses->server.out.data)
			return buf_move(&ses->server.out, &ses->client.in, len);
	}
	else
		return buf_remove(&ses->client.in, len);
	return 0;
}

int client_parse(struct session *ses)
{
	struct connection *client = &ses->client;
	char *start, *lf_pos;
	int len = 0;

	if(!client->in.size)
		return 1; // empty buffer, nothing to parse

	start = client->in.data;

	while(client->in.size)
	{
		dprintf("Parsing client request (size = %i)\n", client->in.size);

		lf_pos = memchr(start, 10, client->in.size); // look for an LF character

		if(client->state != ST_BODY && !lf_pos)
			return 1; // no full line in the buffer yet for METHOD or HEADER states

		if(lf_pos)
			len = lf_pos - start + 1;
		else
			len = client->in.size;

		switch(client->state)
		{
			case ST_METHOD:
				session_handle_method(ses, start, len);
				if(client->method == MET_ERROR)
				{
					session_resp_close_client(client, "501 Not Implemented");
					return 0;
				}
				if(client->method == MET_CONNECT && filter_proxy)
				{
					session_resp_close_client(client, "501 Not Implemented");
					return 0;
				}
				if(buf_move(&ses->server.out, &client->in, len))
					client->state = ST_HEADER;
				else
					return 0;
				break;
			case ST_HEADER:
				if((len == 1 && start[0] == 10) || (len == 2 && start[0] == 13 && start[1] == 10))
				{
					dprintf("Received empty line, waiting for the body\n");
					if(!buf_printf(&ses->server.out, "Connection: close\r\n"))
						return 0;
					if(buf_move(&ses->server.out, &client->in, len))
						client->state = ST_BODY;
					else
						return 0;
				}
				else
				{
					dprintf("Received header line: '");
					dprintbuf(start, len);
					dprintf("'\n");
					if(!client_parse_header(ses, start, len))
						return 0;
				}
				break;
			case ST_BODY:
				if(!buf_move(&ses->server.out, &client->in,
										 min(ses->client.in.size, BUF_SIZE - ses->server.out.size)))
					return 0;
				break;
		}
	}

	return 1;
}

int server_parse_header(struct session *ses, char *hdr, int len)
{
	int copy = 1, slen;

	if(!strncasecmp(hdr, "Set-Cookie: ", min(12, len)) && filter_cookies)
		copy = 0;
	else if(!strncasecmp(hdr, "Keep-Alive: ", min(12, len)))
		copy = 0;
	else if(!strncasecmp(hdr, "Connection: ", min(12, len)))
		copy = 0;
	else if(!strncasecmp(hdr, "Transfer-Encoding: ", min(19, len)))
	{
		slen = len;
		if(hdr[slen-1] == 10)
			slen--;
		if(hdr[slen-1] == 13)
			slen--;
		if(!strncasecmp(hdr+slen-7, "chunked", 7))
		{
			ses->server.chunked = 1;
			dprintf("Chunked!\n");
		}
		copy = 0;
	}
	else if(!strncasecmp(hdr, "Content-Length: ", min(16, len)))
	{
		copy = 0;
		ses->server.cont_len = atoi(hdr+16);
	}
	else if(!strncasecmp(hdr, "Content-Type: ", min(14, len)))
	{
		if(!strncasecmp(hdr+14, "text/", min(5, len-14)))
		{
			ses->server.parsable = 1;
			dprintf("Parsable!\n");
		}
		else if(!strncasecmp(hdr+14, "application/x-javascript", min(24, len-14)))
		{
			ses->server.parsable = 2;
			dprintf("Parsable!\n");
		}
	}

	if(copy)
	{
		if(ses->client.out.data)
			return buf_move(&ses->client.out, &ses->server.in, len);
	}
	else
		return buf_remove(&ses->server.in, len);
	return 0;
}

int server_parse(struct session *ses)
{
	struct connection *server = &ses->server;
	char *start, *lf_pos;
	int len = 0;

	if(!server->in.size)
		return 1; // empty buffer, nothing to parse

	start = server->in.data;

	while(server->in.size)
	{
		dprintf("Parsing server response (size = %i)\n", server->in.size);

		lf_pos = memchr(start, 10, server->in.size); // look for an LF character

		if(server->state != ST_BODY && !lf_pos)
			return 1; // no full line in the buffer yet for STATUS or HEADER states

		if(lf_pos)
			len = lf_pos - start + 1;
		else
			len = server->in.size;

		switch(server->state)
		{
			case ST_STATUS:
				if(buf_move(&ses->client.out, &server->in, len))
					server->state = ST_HEADER;
				else
					return 0;
				break;
			case ST_HEADER:
				if((len == 1 && start[0] == 10) || (len == 2 && start[0] == 13 && start[1] == 10))
				{
					dprintf("Received empty line, dump Content-Length if any and wait for the body\n");
					if(!buf_printf(&ses->client.out, "Connection: close\r\n"))
						return 0;
					if(!ses->server.parsable)
					{
						if(!buf_printf(&ses->client.out, "Content-Length: %d\r\n", ses->server.cont_len))
							return 0;
					}
					if(buf_move(&ses->client.out, &server->in, len))
					{
						server->state = ST_BODY;
						if(ses->server.parsable == 1)
							push_state(server, BST_HTML);
						else if(ses->server.parsable == 2)
							push_state(server, BST_SCRIPT);
						else
							push_state(server, BST_BINARY);
					}
					else
						return 0;
				}
				else
				{
					dprintf("Received header line: '");
					dprintbuf(start, len);
					dprintf("'\n");
					if(!server_parse_header(ses, start, len))
						return 0;
				}
				break;
			case ST_BODY:
				dprintf("ses->client.out.size = %d\n", ses->client.out.size);
				if(ses->server.parsable)
					body_parse_server(ses);
				else
					buf_move(&ses->client.out, &server->in,
									 min(ses->server.in.size, BUF_SIZE - ses->client.out.size));
				return 0;
				break;
		}
	}

	return 1;
}

int session_iterate_delete()
{
	struct session *ptr = session_list, *next;

	while(ptr)
	{
		if(!ptr->client.in.data && !ptr->client.out.size && !ptr->server.in.size)
		{
			dprintf("Client input buffer is not available on ses %d (0x%X)\n", ptr->id, ptr);
			session_close(ptr);
		}

		if(ptr->client.sok < 0 && ptr->server.sok < 0)
		{
			dprintf("Removing session\n");
			next = ptr->next;
			NODE_DEL(session_list, ptr);
			ptr = next;
			continue;
		}
		ptr = ptr->next;
	}
	return 0;
}

int session_iterate_shutdown()
{
	struct session *ptr = session_list, *next;

	while(ptr)
	{
		session_close(ptr);

		dprintf("Removing session\n");
		next = ptr->next;
		NODE_DEL(session_list, ptr);
		ptr = next;
	}
	return 0;
}
