/*******************************************************************************
 * FILE PURPOSE:    SSP Module Device Driver Source.
 *******************************************************************************
 * FILE NAME:       ssp_drv.c
 *
 * DESCRIPTION:     OS abstracted and platform independent SSP module source
  *
 * REVISION HISTORY:
 * 20th Oct 04 - Creation -  TI Germantown 
 *
 * (C) Copyright 2003, Texas Instruments, Inc
 ******************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/spinlock.h> 
#include <pal.h>
#include <asm/avalanche/generic/avalanche_intc.h>

#include "ssp_drv.h"
#include "ssp_hal.h"
#include "ssp_config.h"

#include <asm/semaphore.h>

/* this mutex indicates that the module is in use */
struct semaphore ssp_drvp_module_in_use;

/* this mutex indicates that transfer is done */
struct semaphore ssp_drvp_xfr_complete;

extern SSP_HAL_I2C_INFO_T ssp_i2c_static_config;
extern SSP_HAL_SPI_INFO_T ssp_spi_static_config;
extern SSP_HAL_LCD_CTRL_INFO_T ssp_lcd_ctrl_static_config;

static void ssp_driver_isr(int isr,void *data, struct pt_regs *reg);


/* I2C related stuff */
static devfs_handle_t devfs_dir_handle = NULL;
devfs_handle_t i2c_handle;

static int i2c_open(struct inode * , struct file * );
static int i2c_close(struct inode * , struct file * );
static ssize_t i2c_read( struct file * file, char * buf, size_t count, loff_t *ppos );
static ssize_t i2c_write( struct file * file, const char * buf, size_t count, loff_t *ppos );
static int i2c_ioctl(struct inode *, struct file *,unsigned int , unsigned long );

struct file_operations i2c_fops = 
{
	owner:   THIS_MODULE,
	open :	i2c_open,	
  read :   i2c_read,
  write:   i2c_write,
	ioctl:   i2c_ioctl,
  release : i2c_close,
};

typedef struct 
{
  SSP_HAL_I2C_INFO_T   *id;       

} i2c_dev_t;
  
static i2c_dev_t *i2c_dev_g;

static int i2c_dev_init (void);
static void i2c_dev_exit(void);


/* LCD Control related stuff */
static devfs_handle_t lcd_ctrl_devfs_dir_handle = NULL;
devfs_handle_t lcd_ctrl_handle;

static int lcd_ctrl_open(struct inode * , struct file * );
static int lcd_ctrl_close(struct inode * , struct file * );
static int lcd_ctrl_ioctl(struct inode *, struct file *,unsigned int , unsigned long );

struct file_operations lcd_ctrl_fops = 
{
	owner:   THIS_MODULE,
	open :	lcd_ctrl_open,	
	ioctl:   lcd_ctrl_ioctl,
  release : lcd_ctrl_close,
};

static int lcd_ctrl_dev_init (void);
static void lcd_ctrl_dev_exit(void);


/*******************************************************************************
 * FUNCTION NAME : ssp_drv_cbk_wait_for_xfr_done
 *******************************************************************************
 * DESCRIPTION   : This is the driver callback issued by HAL.
 *
 * RETURNS       : None
 *
 ******************************************************************************/
void ssp_drv_cbk_wait_for_xfr_done()
{
  down( &ssp_drvp_xfr_complete );
}

/*******************************************************************************
 * FUNCTION NAME : ssp_dev_init: Initialize the SSP driver
 *******************************************************************************
 * DESCRIPTION   : The driver init calls the HAL init and connects the 
 *                 interrupt if interrupt mode is used.
 *
 * RETURNS       : SSP_DRV_OK on success, SSP_DRV_ERROR otherwise.
 *
 ******************************************************************************/
int __init ssp_dev_init(void)
{
  /* take module out of reset */
  PAL_sysResetCtrl(AVALANCHE_SSP_RESET_BIT, OUT_OF_RESET);

  /* Initialize HAL */
  if( ssp_hal_init( AVALANCHE_SSP_BASE, PAL_sysClkcGetFreq( CLKC_VBUS ) ) != SSP_DRV_OK )
     return SSP_DRV_ERROR;   
 
  /* Initialize semaphores */
  sema_init( &ssp_drvp_module_in_use,1 );
  sema_init( &ssp_drvp_xfr_complete,0 );
    
  /* Initialize interrupt */
  request_irq( LNXINTNUM( AVALANCHE_SSP_INT ),ssp_driver_isr,0,"ssp",NULL );
  avalanche_intr_type_set( LNXINTNUM( AVALANCHE_SSP_INT ), 0 );
  
  /* Initialize I2C */
  i2c_dev_init();

  /* Initialize LCD control */
  lcd_ctrl_dev_init();
  
  return SSP_DRV_OK;
}

static void ssp_driver_isr(int isr,void *data, struct pt_regs *reg)
{
    /* Disable further interrupts */        
    ssp_hal_intr_ctrl(SSP_HAL_INTR_DISABLE);
    
    /* Acknowledge the current interrupt */
    ssp_hal_intr_ctrl(SSP_HAL_INTR_ACK);

    /* Give the wait semaphore */
    up( &ssp_drvp_xfr_complete );
    
    /* Enable further interrupts */        
    ssp_hal_intr_ctrl(SSP_HAL_INTR_ENABLE);
}

void __exit ssp_dev_exit(void)
{
  i2c_dev_exit();
  lcd_ctrl_dev_exit();
}

/* I2C related functions */

/*******************************************************************************
 * FUNCTION NAME : i2c_dev_init: Initialize the I2C driver
 *******************************************************************************
 * DESCRIPTION   : The driver init function creates i2c device.
 *
 * RETURNS       : SSP_DRV_OK on success, SSP_DRV_ERROR otherwise.
 *
 ******************************************************************************/
static int i2c_dev_init (void)
{
   char devname[10];

   if(PAL_osMemAlloc(0, sizeof(i2c_dev_t), 0, ( void *)&i2c_dev_g) != PAL_SOK) 
   {
     printk( "I2c alloc failed\n" );
     return -EIO;
   }

   i2c_dev_g->id = NULL;
   
   devfs_dir_handle = devfs_mk_dir(NULL, "i2c", NULL);
   sprintf(devname, "0");
   i2c_handle=devfs_register(devfs_dir_handle, devname, DEVFS_FL_AUTO_DEVNUM, 0, 0,
                          S_IFCHR | S_IRUGO | S_IWUGO,&i2c_fops,i2c_dev_g);
 
   return 0;
}

/*******************************************************************************
 * FUNCTION NAME : i2c_open: I2C Driver Open Call
 *******************************************************************************
 * DESCRIPTION   : The I2C Driver open call registers the I2C device with the 
 *                 SSP and returns a descriptor to be used in all future 
 *                 operations on the device.
 *
 * RETURNS       : desc on success, I2C_ERROR otherwise.
 *
 ******************************************************************************/
static int i2c_open(struct inode *i , struct file *file )
{
  i2c_dev_t *i2c_dev = file->private_data;
  SSP_HAL_I2C_INFO_T  info;       
  SSP_DRV_DESC_T      ret;       

  /* Get basic static configuration */
  memcpy(&info, &ssp_i2c_static_config, sizeof(SSP_HAL_I2C_INFO_T));
  
  /* 
   * Take the device mutex 
   */
  if (down_interruptible(&ssp_drvp_module_in_use))
    	return -EBUSY; 

  if((ret = ssp_hal_i2c_open(info)) == NULL){
      return -EFAULT;            
  }
  
  i2c_dev->id = ( void *)ret;

  return 0;
}

/*******************************************************************************
 * FUNCTION NAME : i2c_ioctl: I2C Driver IOCTL Call
 *******************************************************************************
 * DESCRIPTION   : IOCTL calls for the I2C device
 *                 
 * RETURNS       : SSP_DRV_OK for OK, or -EFAULT.
 *
 ******************************************************************************/
static int i2c_ioctl( struct inode * inode, struct file * file,
				             unsigned int cmd, unsigned long arg )
{
   i2c_dev_t *i2c_dev = file->private_data;
   SSP_HAL_I2C_INFO_T *info = i2c_dev->id;
   
   switch ( cmd )
   {
     case I2C_SET_ADDR:
        info->addr = arg;
        break;
        
     default:
        return -EFAULT;
   }
   
   return 0;

}

/*******************************************************************************
 * FUNCTION NAME : i2c_close: I2C Driver Close Call
 *******************************************************************************
 * DESCRIPTION   : Close I2C device 
 *
 * RETURNS       : SSP_DRV_OK on success, SSP_DRV_ERROR otherwise.
 *
 ******************************************************************************/
static int i2c_close(struct inode *i , struct file *file )
{
  i2c_dev_t *i2c_dev = file->private_data;
  
  ssp_hal_i2c_close(i2c_dev->id);

  /*
   * Release device mutex 
   */        
  up(&ssp_drvp_module_in_use);
 
  return 0;
}

/*******************************************************************************
 * FUNCTION NAME : i2c_read: I2C Driver read routine.
 *******************************************************************************
 * DESCRIPTION   : The I2C driver read routine takes the device descriptor
 *                 and performs device read.
 * 
 * RETURNS       : SSP_DRV_OK on success, SSP_DRV_ERROR otherwise.
 *
 ******************************************************************************/
static ssize_t i2c_read( struct file * file, char * buf,
                                                  size_t count, loff_t *ppos )
{
  i2c_dev_t *i2c_dev = file->private_data;
  char p_buff[ I2C_MAX_BUF_SIZE ];   

  if( count > I2C_MAX_BUF_SIZE )
     count = I2C_MAX_BUF_SIZE;

  if( ssp_hal_i2c_read((SSP_HAL_I2C_INFO_T*)i2c_dev->id, p_buff, count) == SSP_HAL_ERROR )
  {
     return -EFAULT;
  }

  if(copy_to_user(buf,p_buff,count))
     return -EFAULT;

  return 0;
}

/*******************************************************************************
 * FUNCTION NAME : i2c_write: I2C Driver write routine.
 *******************************************************************************
 * DESCRIPTION   : The I2C driver read routine takes the device descriptor 
 *                 and performs device read.
 * 
 * RETURNS       : SSP_DRV_OK on success, SSP_DRV_ERROR otherwise.
 *
 ******************************************************************************/
static ssize_t i2c_write( struct file * file, const char * buf,
                                                    size_t count, loff_t *ppos )
{
  i2c_dev_t *i2c_dev = file->private_data;
  char p_buff[ I2C_MAX_BUF_SIZE ];   

  if( count > I2C_MAX_BUF_SIZE )
     count = I2C_MAX_BUF_SIZE;
 
  if(copy_from_user(p_buff,buf,count))
    	return -EFAULT;
  
  if( ssp_hal_i2c_write((SSP_HAL_I2C_INFO_T*)i2c_dev->id, p_buff, count) == SSP_HAL_ERROR )
  {
     return -EFAULT;
  }

  return 0;        
}


/*******************************************************************************
 * FUNCTION NAME : i2c_dev_exit: Removes the I2C driver
 *******************************************************************************
 * DESCRIPTION   : The driver exit function destoys the device.
 *
 * RETURNS       : void.
 *
 ******************************************************************************/
static void i2c_dev_exit(void)
{
	devfs_unregister(i2c_handle);
	devfs_unregister(devfs_dir_handle);
  
  if( i2c_dev_g )
     PAL_osBufFree( 0, ( void *)i2c_dev_g );
}


/*******************************************************************************
 * FUNCTION NAME : lcd_ctrl_dev_init: Initialize the LCD control driver
 *******************************************************************************
 * DESCRIPTION   : The driver init function creates i2c device.
 *
 * RETURNS       : SSP_DRV_OK on success, SSP_DRV_ERROR otherwise.
 *
 ******************************************************************************/
static int lcd_ctrl_dev_init (void)
{
   char devname[10];
  
   lcd_ctrl_devfs_dir_handle = devfs_mk_dir(NULL, "lcd_ctrl", NULL);
   sprintf(devname, "0");
   lcd_ctrl_handle=devfs_register(lcd_ctrl_devfs_dir_handle, devname, DEVFS_FL_AUTO_DEVNUM, 0, 0,
                          S_IFCHR | S_IRUGO | S_IWUGO,&lcd_ctrl_fops,NULL );
 
   return SSP_DRV_OK;
}


/*******************************************************************************
 * FUNCTION NAME : lcd_ctrl_open: LCD control Driver Open Call
 *******************************************************************************
 * DESCRIPTION   : opens LCD control device
 *                 
 * RETURNS       : SSP_DRV_OK.
 *
 ******************************************************************************/
static int lcd_ctrl_open(struct inode *i , struct file *file )
{
  return SSP_DRV_OK;
}

/*******************************************************************************
 * FUNCTION NAME : lcd_ctrl_ioctl: LCD control Driver IOCTL Call
 *******************************************************************************
 * DESCRIPTION   : IOCTL calls for the LCD control device
 *
 * RETURNS       : SSP_DRV_OK for OK, or -EFAULT.
 *
 ******************************************************************************/
static int lcd_ctrl_ioctl( struct inode * inode, struct file * file,
				             unsigned int cmd, unsigned long arg )
{
   SSP_LCD_PULSE_ARG pulse_arg;
   BOOL up;
   int count;
   
   switch ( cmd )
   {
     case LCD_CTRL_CONTRAST:
	    if(copy_from_user(&pulse_arg, (char *)arg, sizeof(SSP_LCD_PULSE_ARG)))
           return -EFAULT;
        
		up = (pulse_arg.is_up)?TRUE:FALSE;
        
        for(count = pulse_arg.cnt; count > 0; count--)
        {
            ssp_hal_lcd_ctrl_send( &ssp_lcd_ctrl_static_config, up );
        }
        break;

     default:
        return -EFAULT;
   }
   
   return SSP_DRV_OK;

}

/*******************************************************************************
 * FUNCTION NAME : lcd_ctrl_close: LCD control Driver Close Call
 *******************************************************************************
 * DESCRIPTION   : Close LCD control device 
 *
 * RETURNS       : SSP_DRV_OK.
 *
 ******************************************************************************/
static int lcd_ctrl_close(struct inode *i , struct file *file )
{
  return SSP_DRV_OK;
}

/*******************************************************************************
 * FUNCTION NAME : lcd_ctrl_dev_exit: Removes the LCD control driver
 *******************************************************************************
 * DESCRIPTION   : The driver exit function destoys the device.
 *
 * RETURNS       : void.
 *
 ******************************************************************************/
static void lcd_ctrl_dev_exit (void)
{
   devfs_unregister(lcd_ctrl_handle);
   devfs_unregister(lcd_ctrl_devfs_dir_handle);
}

module_init(ssp_dev_init);
module_exit(ssp_dev_exit);
