/*
 *  Copyright (C) 2004 Texas Instruments, Inc.  All rights reserved.
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */

/*******************************************************************************
 * FILE PURPOSE:    Graphical LCD Driver and Window Manager Stack Source
 *******************************************************************************
 * FILE NAME:       glcd_hal.c
 *
 * DESCRIPTION:     Graphical LCD HAL Source file
 *
 * REVISION HISTORY:
 * 09-04-2004      Nakshatra     Initial creation
 * 16-02-2005      Hai Shalom
 *
 * (C) Copyright 2004, Texas Instruments, Inc
 ******************************************************************************/

#include "glcd_hal.h"
#include "glcd_pf.h"

UINT32 g_bpp;
UINT32 g_hgt;
UINT32 g_wdh;
UINT32 g_palette_sz;
UINT32 g_bpl;
UINT32 g_databuf_sz;

/* Wait for Interrupt Status to appear. */
GLCD_STATUS LCD_wait_status_int(int int_status_mask)
{
  volatile int counter;

     counter = 0;

  if( int_status_mask == LCD_SYNC_LOST )
  { 
     /* We want to wait for it to clear not be set. */
     while( (LCD_STAT_REG & int_status_mask) && (counter < 100000) )
        counter++;
  }
  else
  {
     while( !(LCD_STAT_REG & int_status_mask) && (counter < 100000) )
        counter++;
  }    
#ifdef FB_DEBUG
  if( counter == 100000 )
  {
     os_print( KERN_ALERT "GLCD: Couldn't find Status Interrupt!\n" );
     LCD_show_raster_status( LCD_STAT_REG );
  }
#endif

  return STATUS_SUCCESS;
}

/* Disable the Raster Engine of the LCD Controller */
GLCD_STATUS LCD_disable_raster(UINT32 wait_for_eof)
{
  LCD_RASTER_CTRL_REG &= ~LCD_RASTER_ENABLE;

  /* Wait for the End of Frame 0 to detect */
  if( wait_for_eof )
  {
     LCD_wait_status_int( LCD_END_OF_FRAME0 );
  }

  return STATUS_SUCCESS;
}

/* Reset the LCD Controller */
void LCD_reset(void)
{
  /* Disable the Raster if previously Enabled */
  if( LCD_RASTER_CTRL_REG & LCD_RASTER_ENABLE )
  {
#ifdef FB_DEBUG
     os_print( KERN_INFO "GLCD: Waiting for Raster Frame Done...\n" );
#endif
     LCD_disable_raster( TRUE );
  }

#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: LCD Raster is going for Reset now...\n" );
  LCD_show_raster_interface( );
#endif

     LCD_RASTER_CTRL_REG = 0;

  /* Place the LCD block in reset */
  SYS_RESET_PERIPH_CTRL_REG &= ~RST_LCD;

  /* Release it from Reset */
  SYS_RESET_PERIPH_CTRL_REG |= RST_LCD;

#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: LCD Controller Reset...\n" );	
  LCD_show_raster_interface( );
#endif
}

/* Configure the Clock Divisor and Set the LCD Mode */
GLCD_STATUS LCD_cfg_clock_divisor(int divisor, int mode)
{
  /* Set the clock divisor, and force LCD_RASTER_MODE. */
  LCD_CTRL_REG = LCD_CLK_DIVISOR( divisor ) | (mode & 0x1);

  return STATUS_SUCCESS;
}

/* Configure the Burst Size of DMA */
GLCD_STATUS LCD_cfg_dma(int burst_size)
{
  UINT32 reg;

  reg = LCD_DMA_CTRL_REG & 0x00000001;
  switch( burst_size )
  {
     case 1:
        reg |= LCD_DMA_BURST_SIZE( LCD_DMA_BURST_1 );
        break;
     case 2:
        reg |= LCD_DMA_BURST_SIZE( LCD_DMA_BURST_2 );
        break;
     case 4:
        reg |= LCD_DMA_BURST_SIZE( LCD_DMA_BURST_4 );
        break;
     case 8:
        reg |= LCD_DMA_BURST_SIZE( LCD_DMA_BURST_8 );
        break;
     case 16:
        reg |= LCD_DMA_BURST_SIZE( LCD_DMA_BURST_16 );
        break;

#ifdef FB_DEBUG  
     os_print( KERN_INFO "GLCD: Configured LCD DMA Burst Size...\n" );
#endif
  }	    
#ifdef EB
  reg |= LCD_DMA_BIG_ENDIAN;
#endif
  LCD_DMA_CTRL_REG = reg | LCD_END_OF_FRAME_INT_ENA;

  return STATUS_SUCCESS;
}

/* Configure the AC Bias */
GLCD_STATUS LCD_cfg_ac_bias(int period, int transitions_per_int)
{
  UINT32 reg;

  /* Set the AC Bias Period and Number of Transisitons per Interrupt */
  reg = LCD_RASTER_TIMING_2_REG & 0xFFF00000;
  reg |= LCD_AC_BIAS_FREQUENCY( period ) | 
     LCD_AC_BIAS_TRANSITIONS_PER_INT( transitions_per_int );

     LCD_RASTER_TIMING_2_REG = reg;                
#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: Configured AC Bias...\n" );
#endif

  return STATUS_SUCCESS;
}

/* Configure the Horizontal Sync Properties */
GLCD_STATUS LCD_cfg_horizontal_sync(int back_porch, int pulse_width, int front_porch)
{
  LCD_RASTER_TIMING_0_REG = ((back_porch & 0xff) << 24)
     | ((front_porch & 0xff) << 16)
     | ((pulse_width & 0x3f) << 10)
     | (LCD_RASTER_TIMING_0_REG & 0xf) ;

#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: Configured Horizontal Sync Properties...\n" );
#endif

  return STATUS_SUCCESS;
}

/* Configure the Vertical Sync Properties */
GLCD_STATUS LCD_cfg_vertical_sync(int back_porch, int pulse_width, int front_porch)
{
  LCD_RASTER_TIMING_1_REG = ((back_porch & 0xff) << 24)
     | ((front_porch & 0xff) << 16)
     | ((pulse_width & 0x3f) << 10)
     | (LCD_RASTER_TIMING_1_REG & 0x3ff);

#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: Configured Vertical Sync Properties...\n" );
#endif

  return STATUS_SUCCESS;
}

GLCD_STATUS LCD_cfg_display(const LCD_CNTRL_CONFIG_T *cfg)
{
  UINT32 reg;

  reg = LCD_RASTER_CTRL_REG & ~(LCD_TFT_MODE | LCD_MONO_8BIT_MODE | LCD_MONOCHROME_MODE);

  switch ( cfg->p_disp_panel->panel_shade )
  {
  case MONOCROME:
     reg |= LCD_MONOCHROME_MODE;
     if( cfg->mono_8bit_mode )
        reg |= LCD_MONO_8BIT_MODE;
     break;
     
  case COLOR_ACTIVE:
     reg |= LCD_TFT_MODE;
     if( cfg->tft_alt_mode )
        reg |= LCD_TFT_ALT_ENABLE;
     break;
     
  case COLOR_PASSIVE:
     if( cfg->stn_565_mode )
        reg |= LCD_STN_565_ENABLE;
     break;   

  default:
#ifdef FB_DEBUG
     os_print( "Undefined LCD type\n" );
#endif
     break;
  }

  LCD_RASTER_CTRL_REG = reg;
  
  reg = LCD_RASTER_TIMING_2_REG;

  if ( cfg->sync_ctrl )
     reg |= LCD_SYNC_CTRL;
  else
     reg &= ~LCD_SYNC_CTRL;

  if ( cfg->sync_edge )
     reg |= LCD_SYNC_EDGE;
  else
     reg &= ~LCD_SYNC_EDGE;
  
  if ( cfg->invert_pxl_clock )
     reg |= LCD_INVERT_PIXEL_CLOCK;
  else
     reg &= ~LCD_INVERT_PIXEL_CLOCK;
  
  if ( cfg->invert_line_clock )
     reg |= LCD_INVERT_LINE_CLOCK;
  else
     reg &= ~LCD_INVERT_LINE_CLOCK;

  if ( cfg->invert_frm_clock )
     reg |= LCD_INVERT_FRAME_CLOCK;
  else
     reg &= ~LCD_INVERT_FRAME_CLOCK;

  LCD_RASTER_TIMING_2_REG = reg;
  
#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: Configured to Active Color...\n" );
#endif

  return STATUS_SUCCESS;
}

/* Configure the Frame Buffer */
GLCD_STATUS LCD_cfg_frame_buffer(UINT32 width, UINT32 height, UINT32 bpp, UINT32 raster_order)
{
  UINT32 reg;
  g_hgt = height;
  g_wdh = width & 0x7f0;
  g_bpp = bpp;

  /* Disable Dual Frame Buffer. */
  LCD_DMA_CTRL_REG = LCD_DMA_CTRL_REG & ~LCD_DUAL_FRAME_BUFFER_ENABLE;

  /* Set the Panel Width */
  LCD_RASTER_TIMING_0_REG = (((g_wdh >> 4) - 1) << 4) | (LCD_RASTER_TIMING_0_REG & 0xfffffc00);

  /* Set the Panel Height */
  LCD_RASTER_TIMING_1_REG = ((g_hgt - 1) & 0x3ff) | (LCD_RASTER_TIMING_1_REG & 0xfffffc00);

  /* Set the Raster Order of the Frame Buffer */
  reg = LCD_RASTER_CTRL_REG & ~(1 << 8);
  if( raster_order )
     reg |= LCD_RASTER_ORDER;

     LCD_RASTER_CTRL_REG = (reg & ~0x300000) | LCD_PALETTE_LOAD_MODE( PALETTE_ONLY );

  switch( g_bpp )
  {
     case 1:
     case 2:
     case 4:
     case 16:
        g_palette_sz = 16 * 2;
        g_bpl = g_wdh * g_bpp / 8;
        g_databuf_sz = g_hgt * g_bpl;
        break;

     case 8:    
        g_palette_sz = 256 * 2;
        g_bpl = g_wdh * g_bpp / 8;
        g_databuf_sz = g_hgt * g_bpl;
        break;

     default:
#ifdef FB_DEBUG
        os_print( KERN_ALERT "GLCD: Unsupported BPP!\n" );
#endif
        break;
  }
#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: Configured Frame Buffer...\n" );
#endif

  return STATUS_SUCCESS;
}

/* Clear the Interrupt Status Flag */
GLCD_STATUS LCD_clear_status_int(int int_status_mask)
{
  volatile int counter;
  UINT32 reg1, reg2;

  reg1 = LCD_STAT_REG;

  counter = 0;

  /* Clear interrupt and wait a while for it to 'take'... */
  while( (LCD_STAT_REG & int_status_mask) && (counter < 100000) )
  {	
     LCD_STAT_REG = int_status_mask;
     counter++;        
  }

  reg2 = LCD_STAT_REG;

#ifdef FB_DEBUG  
  if( counter == 100000 )
  {
     os_print( KERN_ALERT "GLCD: Couldn't Clear Status Interrupt!\n" );
     os_print( KERN_INFO "GLCD: Status Register Before Clear: 0x%x\n",reg1 );
     os_print( KERN_INFO "GLCD: Status Register After Clear: 0x%x\n",reg2 );
  }
#endif

  return STATUS_SUCCESS;
}

#ifdef FB_DEBUG
/* Show the Raster Status */
void LCD_show_raster_status(UINT32 value)
{
  if( value & LCD_FIFO_UNDERFLOW )
     os_print( KERN_ALERT "GLCD: FIFO Underflow!\n" );

  if( value & LCD_AC_BIAS_COUNT_STATUS )
     os_print( KERN_INFO "GLCD: AC Bias Count Reached\n" );

  if( value & LCD_PALETTE_LOADED )
     os_print( KERN_INFO "GLCD: Palette is Loaded\n" );
  else
     os_print( KERN_ALERT "GLCD: Palette is not Loaded!\n" );

  if( value & LCD_SYNC_LOST )
  {
     os_print( KERN_ALERT "GLCD: Frame Synchronization Error!\n" );
     LCD_STAT_REG = LCD_STAT_REG | LCD_SYNC_LOST;
  }
}
#endif

#ifdef FB_DEBUG
/* Show the Raster Interface */
void LCD_show_raster_interface(void)
{
  os_print( KERN_INFO "\n" );    
  os_print( KERN_INFO "LCD Raster Interface Configuration\n" );
  os_print( KERN_INFO "----------------------------------\n" );
  os_print( KERN_INFO "LCD Control Register:	0x%x\n", LCD_CTRL_REG );
  os_print( KERN_INFO "Raster Control:       	0x%x\n", LCD_RASTER_CTRL_REG );
  os_print( KERN_INFO "Timing Register 0:    	0x%x\n", LCD_RASTER_TIMING_0_REG );
  os_print( KERN_INFO "Timing Register 1:    	0x%x\n", LCD_RASTER_TIMING_1_REG );
  os_print( KERN_INFO "Timing Register 2:    	0x%x\n", LCD_RASTER_TIMING_2_REG );
  os_print( KERN_INFO "Subpanel Register:	0x%x\n", LCD_RASTER_SUBPANEL_DISP_REG );
  os_print( KERN_INFO "DMA Control Register:	0x%x\n", LCD_DMA_CTRL_REG );
  os_print( KERN_INFO "DMA 0 Start Address:	0x%x\n", LCD_DMA_FRM_BUF_BASE_ADDR_0_REG );
  os_print( KERN_INFO "DMA 0 End Address:	0x%x\n", LCD_DMA_FRM_BUF_CEILING_ADDR_0_REG );
  os_print( KERN_INFO "DMA 1 Start Address:	0x%x\n", LCD_DMA_FRM_BUF_BASE_ADDR_1_REG );
  os_print( KERN_INFO "DMA 1 End Address:	0x%x\n", LCD_DMA_FRM_BUF_CEILING_ADDR_1_REG );
  os_print( KERN_INFO "\n" );
  LCD_show_raster_status( LCD_STAT_REG );
  os_print( KERN_INFO "\n" );

}
#endif

/* Initialize the GPIO Pins of the LCD Controller */
void lcd_tb_gpio_init( void )
{
  volatile UINT32 reg;

  /* Enable the LCD Pins */
  reg = SYS_RESET_PIN_SEL_REG( 8 );
  reg	|= 0x44000000;
  SYS_RESET_PIN_SEL_REG( 8 ) = reg;

  reg = SYS_RESET_PIN_SEL_REG( 9 );
  reg	|= 0x55145515;
  SYS_RESET_PIN_SEL_REG( 9 ) = reg;

  reg = SYS_RESET_PIN_SEL_REG( 10 );
  reg	|= 0x00005505;
  SYS_RESET_PIN_SEL_REG( 10 ) = reg;

  /* Enable the GPIO pin 6 for VDD_LCD Control
     GPIO6 pin has GPIO as primary functionality, by default it is enabled
     It is set as Output Port and the default Output value is Zero */

  /* Set the GPIO6 Pin */
  reg = SYS_GPIO_DATA_OUT_0_REG;
  reg |= (1 << 6);
  SYS_GPIO_DATA_OUT_0_REG = reg;

#ifdef FB_DEBUG
  os_print( KERN_INFO "GLCD: GPIO Pins Intialized\n" );
#endif
}
