/*
 *	linux/kernel/softirq.c
 *
 *	Copyright (C) 1992 Linus Torvalds
 *
 * Fixed a disable_bh()/enable_bh() race (was causing a console lockup)
 * due bh_mask_count not atomic handling. Copyright (C) 1998  Andrea Arcangeli
 *
 * Rewritten. Old one was good in 2.2, but in 2.3 it was immoral. --ANK (990903)
 */

//-------------------------------------------------------------------------------------
// Copyright 2005, Texas Instruments Incorporated
//
// This program has been modified from its original operation by Texas Instruments
// to do the following:
// 
// 1. Voice-Data coexistence scheduling fixes.
//
// THIS MODIFIED SOFTWARE AND DOCUMENTATION ARE PROVIDED
// "AS IS," AND TEXAS INSTRUMENTS MAKES NO REPRESENTATIONS 
// OR WARRENTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
// TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY 
// PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR 
// DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS, 
// COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. 
// See The GNU General Public License for more details.
//
// These changes are covered under version 2 of the GNU General Public License, 
// dated June 1991.
//-------------------------------------------------------------------------------------

#include <linux/config.h>
#include <linux/mm.h>
#include <linux/kernel_stat.h>
#include <linux/interrupt.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/tqueue.h>

#include <linux/trace.h>

/* These define the two level of priorities that are present in the System:-
 *  1) Timer and Hi Softirq's are still executed immediately after interrupts.
 *  2) All Other SOFTIRQ's are treated as low priority and are executed only
 *     in the daemon context.*/
#define NSP_ALWAYS_SOFTIRQ      ((1 << RUN_TIMER_LIST) | (1 << HI_SOFTIRQ))
#define NSP_ALWAYS_NON_SOFTIRQ  (~NSP_ALWAYS_SOFTIRQ)

/*
   - No shared variables, all the data are CPU local.
   - If a softirq needs serialization, let it serialize itself
     by its own spinlocks.
   - Even if softirq is serialized, only local cpu is marked for
     execution. Hence, we get something sort of weak cpu binding.
     Though it is still not clear, will it result in better locality
     or will not.
   - These softirqs are not masked by global cli() and start_bh_atomic()
     (by clear reasons). Hence, old parts of code still using global locks
     MUST NOT use softirqs, but insert interfacing routines acquiring
     global locks. F.e. look at BHs implementation.

   Examples:
   - NET RX softirq. It is multithreaded and does not require
     any global serialization.
   - NET TX softirq. It kicks software netdevice queues, hence
     it is logically serialized per device, but this serialization
     is invisible to common code.
   - Tasklets: serialized wrt itself.
   - Bottom halves: globally serialized, grr...
 */

irq_cpustat_t irq_stat[NR_CPUS];

static struct softirq_action softirq_vec[32] __cacheline_aligned;

#ifdef CONFIG_PREEMPT_TIMES
extern void latency_cause(int,int);
#else
#define latency_cause(a, b)
#endif

extern int gIsVoiceEnabled;

void trigger_softirq(void)
{
    if(softirq_pending(smp_processor_id())) 
	    do_softirq();
}
        
/*
 * we cannot loop indefinitely here to avoid userspace starvation,
 * but we also don't want to introduce a worst case 1/HZ latency
 * to the pending events, so lets the scheduler to balance
 * the softirq load for us.
 */
static inline void wakeup_softirqd(unsigned cpu)
{
	struct task_struct * tsk = ksoftirqd_task(cpu);

	if (tsk && tsk->state != TASK_RUNNING)
		wake_up_process(tsk);
}

/**************************************************************************
 * FUNCTION NAME : do_softirq_original
 **************************************************************************
 * DESCRIPTION   :
 *  The original LINUX 'do_softirq' code. This function was renamed because the
 *  function do_softirq was a generic API interface which gets called from 
 *  many places in the System to execute the bottom halves. This is called
 *  only when voice is disabled.
 ***************************************************************************/
asmlinkage void do_softirq_original()
{
	int cpu = smp_processor_id();
	__u32 pending;
	long flags;
	__u32 mask;

	if (in_interrupt())
		return;
	preempt_lock_start(-99);

	local_irq_save(flags);

	pending = softirq_pending(cpu);

	if (pending) {
		struct softirq_action *h;

		mask = ~pending;

		local_bh_disable();
restart:
		/* Reset the pending bitmask before enabling irqs */
		softirq_pending(cpu) = 0;

		local_irq_enable();

		h = softirq_vec;
 
		latency_cause(-100, pending);


		do {
			if (pending & 1) {
		                TRACE_SOFT_IRQ(TRACE_EV_SOFT_IRQ_SOFT_IRQ, (h - softirq_vec));
				h->action(h);
			}
			h++;
			pending >>= 1;
		} while (pending);

		local_irq_disable();

		pending = softirq_pending(cpu);

		if (pending & (mask | ALWAYS_SOFTIRQ)) {
			mask &= ~pending;
			goto restart;
		}

		__local_bh_enable();

		if (pending)
			wakeup_softirqd(cpu);
	}

	local_irq_restore(flags);
#ifdef CONFIG_ILATENCY
	/*
	 * If you have entry.S whacked up to call intr_ret_from_exception
	 * you can remove this call, unfortunately, MIPs and ARM don't
	 * have a simple way to do this for now so we catch them all here . . 
	 */
	intr_ret_from_exception();
#endif 
}

/**************************************************************************
 * FUNCTION NAME : do_softirq_voice_low
 **************************************************************************
 * DESCRIPTION   :
 *  This function is called to do low priority software BH's. This function
 *  is called from the ksoftirq daemon.
 ***************************************************************************/
static void do_softirq_voice_low()
{
	int cpu = smp_processor_id();
	__u32 pending;
	long flags;

    /* Sanity Check: This function can be called from the context of the ksoftirqd daemon. */
    if (current->pid != 3)
        printk ("Error: Calling the low priority softirq from process:%d.\n", current->pid);

	preempt_lock_start(-99);

    /* Disable interrupts. */
	local_irq_save(flags);

    /* Take a snapshot of pending softirq's. */
	pending = softirq_pending(cpu) & NSP_ALWAYS_NON_SOFTIRQ;
    softirq_pending(cpu) &= ~NSP_ALWAYS_NON_SOFTIRQ;

    /* Re-enable interrupts. */
    local_irq_enable();

    /* Process only NON-ALWAYS softirq. */
	if (pending) 
    {
		struct softirq_action *h;

        /* Cycle through the pending non immediate softirq's and process them. */
		h = softirq_vec;
    	latency_cause(-100, pending);
		do 
        {
			if (pending & 1) 
            {
		        TRACE_SOFT_IRQ(TRACE_EV_SOFT_IRQ_SOFT_IRQ, (h - softirq_vec));
                if (atomic_inc_return(&h->in_use) == 1) 
                {
				    h->action(h);
                }
                else
                {
                    printk("ERROR: Nested softirqs happening from daemon. %ld, 0x%p\n",
                        ((unsigned long)h-(unsigned long)softirq_vec)/sizeof(*h), h->action);
                }
                atomic_dec(&h->in_use);
			}
			h++;
			pending >>= 1;
		} while (pending);
	}

    /* Re-enable interrupts. */
	local_irq_restore(flags);
#ifdef CONFIG_ILATENCY
	/*
	 * If you have entry.S whacked up to call intr_ret_from_exception
	 * you can remove this call, unfortunately, MIPs and ARM don't
	 * have a simple way to do this for now so we catch them all here . . 
	 */
	intr_ret_from_exception();
#endif 
}

/**************************************************************************
 * FUNCTION NAME : do_softirq_voice_high
 **************************************************************************
 * DESCRIPTION   :
 *  This function gets called immediately after interrupt context and its 
 *  duty is to execute the NSP_ALWAYS_SOFTIRQ (Voice Hi Tasklet and Timer BH) 
 *  immediately. Rest of the tasklets are scheduled and executed in the 
 *  context of the softirq daemon. 
 ***************************************************************************/
static void do_softirq_voice_high()
{
	int cpu = smp_processor_id();
	__u32 pending;
	long flags;
	__u32 mask;
    int     counter = 0;

	if (in_interrupt())
		return;
	preempt_lock_start(-99);

    /* Disable interrupts. */
	local_irq_save(flags);

    /* Take a snapshot of all pending software interrupts. */
	pending = softirq_pending(cpu);

    /* Verify if there are any IMMEDIATE softirq's available for immediate attention. 
     * These would be the VOICE HI-TASKLET and the System Timer BH. */
	if (pending & NSP_ALWAYS_SOFTIRQ)
    {
		struct softirq_action *h;

        /* Disable Bottom Halve's from interrupting the processing. */
		local_bh_disable();

        /* Compute the mask. */
        mask = ~(NSP_ALWAYS_SOFTIRQ);

restart:
        /* Clean all but the NOT always immediate softirq's. */
        softirq_pending(cpu) = pending & mask;

        /* Process only the immediate the softirq's. */
        pending = pending & NSP_ALWAYS_SOFTIRQ;

        /* Reenable interrupts. */
		local_irq_enable();

        /* Cycle through the pending softirq's and process them. */
		h = softirq_vec;
		latency_cause(-100, pending);
		do 
        {
			if (pending & 1) 
            {
		        TRACE_SOFT_IRQ(TRACE_EV_SOFT_IRQ_SOFT_IRQ, (h - softirq_vec));
                if (atomic_inc_return(&h->in_use) == 1) 
                {
				    h->action(h);
                }
                else
                {
                    printk("ERROR: Nested softirqs happening from interrupt. %ld, 0x%p\n",
                        ((unsigned long)h-(unsigned long)softirq_vec)/sizeof(*h), h->action);
                }
                atomic_dec(&h->in_use);
			}
			h++;
			pending >>= 1;
		} while (pending);

		local_irq_disable();

        /* Check if there are any pending softirq's which occured while processing the previous
         * snapshot. */
		pending = softirq_pending(cpu);

#if 0
        /* Debug Messages. */
        if (counter > 2)
            printk ("DEBUG: Restarting has already been done %d times..\n", counter);
#endif

        /* If there were any more IMMEDIATE softirq's pending process them immediately. Typically
         * this should not happen a lot. If this happens the the handler's for all immediate BH's
         * need to be investigated. Remember this function runs immediately after IRQ's and will
         * kill the system scheduler if we loop around this function for a long time. */
		if (pending & NSP_ALWAYS_SOFTIRQ)
        {
            counter++;
			goto restart;
		}

        /* Re-enable bottom halves, since we are done with our work. */
		__local_bh_enable();

        /* This implies that all the immediate BH's are taken care of. If there are more non-important
         * softirqs pending, wake up the daemon to handle those. */
		if (pending)
			wakeup_softirqd(cpu);
	}
    else
    {
        /* NOT one of the immediate softirq's. Wake up the daemon. */
			wakeup_softirqd(cpu);
    }

	local_irq_restore(flags);
#ifdef CONFIG_ILATENCY
	/*
	 * If you have entry.S whacked up to call intr_ret_from_exception
	 * you can remove this call, unfortunately, MIPs and ARM don't
	 * have a simple way to do this for now so we catch them all here . . 
	 */
	intr_ret_from_exception();
#endif 
}

/**************************************************************************
 * FUNCTION NAME : do_softirq_from_interrupt
 **************************************************************************
 * DESCRIPTION   :
 *  This is the function that is called from the interrupt handlers after
 *  processing the function. 
 ***************************************************************************/
void asmlinkage do_softirq_from_interrupt()
{
    /* Check if voice mode is enabled or not? */
    if (gIsVoiceEnabled == 0)
    {
        /* Voice Mode Disabled: Do the original code. */
        do_softirq_original();
    }
    else
    {
        /* Voice Mode Enabled: Do only the HIGH priority bottom halves. */
        do_softirq_voice_high();
    }
    return;
}

/**************************************************************************
 * FUNCTION NAME : do_softirq
 **************************************************************************
 * DESCRIPTION   :
 *  The generic well known API to execute BH's. This function calls the 
 *  original implementation if NO Voice call is ACTIVE. In Voice mode this 
 *  call does nothing.
 ***************************************************************************/
asmlinkage void do_softirq()
{
    /* Is Voice Mode off. Call the original implementation. */
    if (gIsVoiceEnabled == 0)
    {
        do_softirq_original();
    } else {
        /* Nothing done here for voice mode. Rely on do_softirq_from_interrupt() and softirqd */
    }
}

/*
 * This function must run with irq disabled!
 */
inline void cpu_raise_softirq(unsigned int cpu, unsigned int nr)
{
	__cpu_raise_softirq(cpu, nr);

	/*
	 * If we're in an interrupt or bh, we're done
	 * (this also catches bh-disabled code). We will
	 * actually run the softirq once we return from
	 * the irq or bh.
	 *
	 * Otherwise we wake up ksoftirqd to make sure we
	 * schedule the softirq soon.
	 */
	if (!(local_irq_count(cpu) | local_bh_count(cpu)))
		wakeup_softirqd(cpu);
}

void raise_softirq(unsigned int nr)
{
	long flags;

	local_irq_save(flags);
	cpu_raise_softirq(smp_processor_id(), nr);
	local_irq_restore(flags);
}

void open_softirq(int nr, void (*action)(struct softirq_action*), void *data)
{
	softirq_vec[nr].data = data;
	softirq_vec[nr].action = action;
    atomic_set(&softirq_vec[nr].in_use,0);
}


/* Tasklets */

struct tasklet_head tasklet_vec[NR_CPUS] __cacheline_aligned;
struct tasklet_head tasklet_hi_vec[NR_CPUS] __cacheline_aligned;

void __tasklet_schedule(struct tasklet_struct *t)
{
	int cpu = smp_processor_id();
	unsigned long flags;

	local_irq_save(flags);
	t->next = tasklet_vec[cpu].list;
	tasklet_vec[cpu].list = t;
	cpu_raise_softirq(cpu, TASKLET_SOFTIRQ);
	local_irq_restore(flags);
}

void __tasklet_hi_schedule(struct tasklet_struct *t)
{
	int cpu = smp_processor_id();
	unsigned long flags;

	local_irq_save(flags);
	t->next = tasklet_hi_vec[cpu].list;
	tasklet_hi_vec[cpu].list = t;
	cpu_raise_softirq(cpu, HI_SOFTIRQ);
	local_irq_restore(flags);
}

static void tasklet_action(struct softirq_action *a)
{
	int cpu = smp_processor_id();
	struct tasklet_struct *list;

	local_irq_disable();
	list = tasklet_vec[cpu].list;
	tasklet_vec[cpu].list = NULL;
	local_irq_enable();

	while (list) {
		struct tasklet_struct *t = list;

		list = list->next;

		if (tasklet_trylock(t)) {
			if (!atomic_read(&t->count)) {
				if (!test_and_clear_bit(TASKLET_STATE_SCHED, &t->state))
					BUG();

				TRACE_SOFT_IRQ(TRACE_EV_SOFT_IRQ_TASKLET_ACTION, (unsigned long) (t->func));

				t->func(t->data);
				tasklet_unlock(t);
				continue;
			}
			tasklet_unlock(t);
		}

		local_irq_disable();
		t->next = tasklet_vec[cpu].list;
		tasklet_vec[cpu].list = t;
		__cpu_raise_softirq(cpu, TASKLET_SOFTIRQ);
		local_irq_enable();
	}
}

static void tasklet_hi_action(struct softirq_action *a)
{
	int cpu = smp_processor_id();
	struct tasklet_struct *list;

	local_irq_disable();
	list = tasklet_hi_vec[cpu].list;
	tasklet_hi_vec[cpu].list = NULL;
	local_irq_enable();

	while (list) {
		struct tasklet_struct *t = list;

		list = list->next;

		if (tasklet_trylock(t)) {
			if (!atomic_read(&t->count)) {
				if (!test_and_clear_bit(TASKLET_STATE_SCHED, &t->state))
					BUG();

				TRACE_SOFT_IRQ(TRACE_EV_SOFT_IRQ_TASKLET_HI_ACTION, (unsigned long) (t->func));

				t->func(t->data);
				tasklet_unlock(t);
				continue;
			}
			tasklet_unlock(t);
		}

		local_irq_disable();
		t->next = tasklet_hi_vec[cpu].list;
		tasklet_hi_vec[cpu].list = t;
		__cpu_raise_softirq(cpu, HI_SOFTIRQ);
		local_irq_enable();
	}
}


void tasklet_init(struct tasklet_struct *t,
		  void (*func)(unsigned long), unsigned long data)
{
	t->next = NULL;
	t->state = 0;
	atomic_set(&t->count, 0);
	t->func = func;
	t->data = data;
}

void tasklet_kill(struct tasklet_struct *t)
{
	if (in_interrupt())
		printk("Attempt to kill tasklet from interrupt\n");

	while (test_and_set_bit(TASKLET_STATE_SCHED, &t->state)) {
		current->state = TASK_RUNNING;
		do {
			current->policy |= SCHED_YIELD;
			schedule();
		} while (test_bit(TASKLET_STATE_SCHED, &t->state));
	}
	tasklet_unlock_wait(t);
	clear_bit(TASKLET_STATE_SCHED, &t->state);
}



/* Old style BHs */

static void (*bh_base[32])(void);
struct tasklet_struct bh_task_vec[32];

/* BHs are serialized by spinlock global_bh_lock.

   It is still possible to make synchronize_bh() as
   spin_unlock_wait(&global_bh_lock). This operation is not used
   by kernel now, so that this lock is not made private only
   due to wait_on_irq().

   It can be removed only after auditing all the BHs.
 */
spinlock_t global_bh_lock = SPIN_LOCK_UNLOCKED;

static void bh_action(unsigned long nr)
{
	int cpu = smp_processor_id();

	if (!spin_trylock(&global_bh_lock))
		goto resched;

	if (!hardirq_trylock(cpu))
		goto resched_unlock;

	if (bh_base[nr]){
	        TRACE_SOFT_IRQ(TRACE_EV_SOFT_IRQ_BOTTOM_HALF, (nr));
		bh_base[nr]();
	}

	hardirq_endlock(cpu);
	spin_unlock(&global_bh_lock);
	return;

resched_unlock:
	spin_unlock(&global_bh_lock);
resched:
	mark_bh(nr);
}

void init_bh(int nr, void (*routine)(void))
{
	bh_base[nr] = routine;
	mb();
}

void remove_bh(int nr)
{
	tasklet_kill(bh_task_vec+nr);
	bh_base[nr] = NULL;
}

void __init softirq_init()
{
	int i;

	for (i=0; i<32; i++)
		tasklet_init(bh_task_vec+i, bh_action, i);

	open_softirq(TASKLET_SOFTIRQ, tasklet_action, NULL);
	open_softirq(HI_SOFTIRQ, tasklet_hi_action, NULL);
        open_softirq(RUN_TIMER_LIST,timer_softirq, NULL);
}

void __run_task_queue(task_queue *list)
{
	struct list_head head, *next;
	unsigned long flags;

	spin_lock_irqsave(&tqueue_lock, flags);
	list_add(&head, list);
	list_del_init(list);
	spin_unlock_irqrestore(&tqueue_lock, flags);

	next = head.next;
	while (next != &head) {
		void (*f) (void *);
		struct tq_struct *p;
		void *data;

		p = list_entry(next, struct tq_struct, list);
		next = next->next;
		f = p->routine;
		data = p->data;
		wmb();
		p->sync = 0;
		if (f)
			f(data);
	}
}

/**************************************************************************
 * FUNCTION NAME : original_process_soft_irq 
 **************************************************************************
 * DESCRIPTION   :
 *  This is the original soft irq processing loop. In the original code we 
 *  handle all the softirq's immediately.
 ***************************************************************************/
void original_process_soft_irq (int cpu)
{
		if (!softirq_pending(cpu))
			schedule();

		__set_current_state(TASK_RUNNING);

	while (softirq_pending(cpu)) 
    {
        /* Voice Mode Disabled: Do the original code. */
        do_softirq_original();

			if (current->need_resched)
				schedule();
		}

		__set_current_state(TASK_INTERRUPTIBLE);
    return;
}

/**************************************************************************
 * FUNCTION NAME : voice_process_soft_irq
 **************************************************************************
 * DESCRIPTION   :
 *  Dispatch function called from the daemon to handle PENDING soft irq's.
 *  Since we handle only NON immediate SOFT-IRQ's we check only if they are
 *  pending.
 ***************************************************************************/
void voice_process_soft_irq (int cpu)
{
    /* If there are no immediate soft irq's pending, call the kernel scheduler. */
    if ((softirq_pending(cpu) & NSP_ALWAYS_NON_SOFTIRQ) == 0)
	    schedule();

    __set_current_state(TASK_RUNNING);

    /* Cycle through the loop only if there are non-immediate soft irqs pending. */
	while ((softirq_pending(cpu) & NSP_ALWAYS_NON_SOFTIRQ) != 0)
    {
        /* Voice Mode Enable: Do the low priority soft irq's */
        do_softirq_voice_low();

        /* Is there a high priority soft irq pending. */
        if (softirq_pending(cpu) & NSP_ALWAYS_SOFTIRQ)
            do_softirq_voice_high();

        /* Check if the process needs to be rescheduled. If so call the scheduler. */
	    if (current->need_resched)
            schedule();
    }
    __set_current_state(TASK_INTERRUPTIBLE);
    return;
}

static int ksoftirqd_priority = -1, ksoftirqd_policy = -1;

void ksoftirqd_high_prio(int cpu) {
    if (ksoftirqd_priority == -1) return;
    ksoftirqd_task(cpu)->policy = SCHED_FIFO;
    ksoftirqd_task(cpu)->rt_priority = 89;
}

void ksoftirqd_norm_prio(int cpu) {
    if (ksoftirqd_priority == -1) return;
    ksoftirqd_task(cpu)->policy = ksoftirqd_policy;
    ksoftirqd_task(cpu)->rt_priority = ksoftirqd_priority;
    ksoftirqd_task(cpu)->nice = 19;
}

static int ksoftirqd(void * __bind_cpu)
{
	int bind_cpu = (int) (long) __bind_cpu;
	int cpu = cpu_logical_map(bind_cpu);

	daemonize();
	current->nice = 19;
	sigfillset(&current->blocked);

	/* Migrate to the right CPU */
	current->cpus_allowed = 1UL << cpu;
	while (smp_processor_id() != cpu)
		schedule();

	sprintf(current->comm, "ksoftirqd_CPU%d", bind_cpu);

    ksoftirqd_policy = current->policy;
    ksoftirqd_priority = current->rt_priority;

    printk ("DEBUG: Starting ksoftirqd with priority:%d policy:%d nice:%d.\n",ksoftirqd_priority, ksoftirqd_policy, current->nice);

	__set_current_state(TASK_INTERRUPTIBLE);
	mb();

	ksoftirqd_task(cpu) = current;

	for (;;) 
    {
        if (gIsVoiceEnabled == 0)
            original_process_soft_irq (cpu);
        else
            voice_process_soft_irq (cpu);
	}
}

static __init int spawn_ksoftirqd(void)
{
	int cpu;

    /* PANKAJ: Test Code. */
    printk ("DEBUG: Starting ksoftirqd with ALWAYS_SOFTIRQ:0x%x NON ALWAYS SOFTIRQ:0x%x.\n", 
            NSP_ALWAYS_SOFTIRQ, NSP_ALWAYS_NON_SOFTIRQ); 

	for (cpu = 0; cpu < smp_num_cpus; cpu++) 
    {
		if (kernel_thread(ksoftirqd, (void *) (long) cpu, CLONE_FS | CLONE_FILES | CLONE_SIGNAL) < 0)
			printk("spawn_ksoftirqd() failed for cpu %d\n", cpu);
		else 
        {
			while (!ksoftirqd_task(cpu_logical_map(cpu))) 
            {
				current->policy |= SCHED_YIELD;
				schedule();
			}
		}
	}

	return 0;
}

__initcall(spawn_ksoftirqd);
