/*
 *	Handle incoming frames
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_input.c,v 1.5 2004/12/27 11:24:54 gopal_kr Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

//-------------------------------------------------------------------------------------
// Copyright 2004, Texas Instruments Incorporated
//
// This program has been modified from its original operation by Texas Instruments
// to do the following:
// 
// 1. Performance optimizations.
// 2. Instrumentation.
//
// THIS MODIFIED SOFTWARE AND DOCUMENTATION ARE PROVIDED
// "AS IS," AND TEXAS INSTRUMENTS MAKES NO REPRESENTATIONS 
// OR WARRENTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
// TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY 
// PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR 
// DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS, 
// COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. 
// See The GNU General Public License for more details.
//
// These changes are covered under version 2 of the GNU General Public License, 
// dated June 1991.
//-------------------------------------------------------------------------------------

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_bridge.h>
#include <linux/netfilter_bridge.h>
#include "br_private.h"
#include "br_mfd.h"

#include <asm/uaccess.h>

/* Instrumentation Support. */
#include <linux/psp_trace.h>

unsigned char bridge_ula[6] = { 0x01, 0x80, 0xc2, 0x00, 0x00, 0x00 };

/*************************** EXTERN DECLARATIONS ***********************/
extern void br_flood(struct net_bridge *br, struct sk_buff *skb, int clone);
extern void br_send_frame(struct net_bridge_port *to, struct sk_buff *skb);

/*************************** STATIC DECLARATIONS ***********************/
static void br_input_error(struct sk_buff* skb);

/**************************************************************************
 * FUNCTION NAME : br_pass_frame_up
 **************************************************************************
 * DESCRIPTION   :
 *  The function passes the received frame up the IP Stack through the
 *  bridge interface.
***************************************************************************/
void br_pass_frame_up(struct net_bridge *br, struct sk_buff *skb)
{
    /* Increment the statistics. */
	br->statistics.rx_packets++;
	br->statistics.rx_bytes += skb->len;

	DEBUGP("br_pass_frame_up from %s\n", skb->dev->name);	

    /* Initialize the device and push the frame up the stack. */
	skb->dev      = &br->dev;
	skb->pkt_type = PACKET_HOST;

    /* TODO: Do we need to do this? Simply setting the protocol field should 
     * sufficient? There is a memcmp in the function which needs to be avoided.
     * We have already made sure that the packet is for us. */
	skb->protocol = eth_type_trans(skb, &br->dev);

    /* Pass packet to the IP stack.*/
    netif_rx(skb);
    return;
}

/**************************************************************************
 * FUNCTION NAME : __br_handle_frame
 **************************************************************************
 * DESCRIPTION   :
 *  The function receives and does the bridge processing. The function
 *  is called with the bridge locked.  
***************************************************************************/
static void __br_handle_frame(struct sk_buff* skb, struct net_bridge* br)
{	
	register unsigned char*                 dest;
	register struct net_bridge_fdb_entry*   fdb_entry;
	register struct net_bridge_port*        source_port;
	int                                     i;
	struct sk_buff*                         skb2;

   /* Initialize the destination address and source port on which the packet arrived. */
	dest        = skb->data;
	source_port = skb->dev->br_port;

    /* Basic Validations. Do not process any packets whose mac address has the highest order
     * byte set. */
	if (skb->data[6] & 1)
	{
		br_input_error(skb);
	        return;
	}

#if 0
    /* Spanning Tree Protocol. */
	if (br->stp_enabled && !memcmp(dest, bridge_ula, 5) && !(dest[5] & 0xF0))
    {
	    if (!dest[5]) 
        {
		    br_stp_handle_bpdu(skb);
    		return;
	    }
    }
#endif

    /* Learn the source mac address only if in the correct state. */
	if (source_port->state & (BR_STATE_LEARNING | BR_STATE_FORWARDING))
		br_fdb_insert(br, source_port, &skb->data[6], 0);

    /* Do not forward if we the port is not in the correct state. */
	if (source_port->state & BR_STATE_FORWARDING)
    {
    	/* NSP Multicast Support: TXTL_IGMP01 20-Oct-2004
	     *  Ethernet frames have a 48 bit destination address field. The IANA reserved a range of
    	 *  addresses for multicast: every ethernet frame with its destination in the range 
    	 *  01-00-5e-00-00-00 to 01-00-5e-ff-ff-ff (hex) contains data for a multicast group. 
	     *  The prefix 01-00-5e identifies the frame as multicast and only 23 bits are left to
    	 *  the multicast address. */
	    if(dest[0] & 1)
    	{
	    	if((!dest[1] && (dest[2] & 0x5e)) && (br->mc_forwarding == MULTICAST_FORWARDING_ENABLED))
		    {
    			DEBUGP("\nbr_handle_frame : rxed multicast from %s\n", skb->dev->name);
    			skb2 = skb_clone(skb, GFP_ATOMIC);
	    		if(skb2)
		    	{
			    	if( br_handle_mc(br, skb2, 1) < 0)	
				    {
    					kfree_skb(skb2);
	    				br_input_error(skb);
		    			return;
			    	}
    			}
	    		else
		    	{
			    	printk(KERN_ERR "In __br_handle_frame : skb_copy failed for multicast data pkt\n");						
				    br_input_error(skb);
    				return;
	    		}
		    	br_pass_frame_up(br,skb);
    		}
	    	else /* Broadcast packet.. flood it*/
		    {	
    			br_flood(br,skb,1);
	    		br_pass_frame_up(br,skb);
		    }
		    return;
	    }

        /* Check the FDB table for a match. */
	    fdb_entry = br_fdb_get(br, dest);
    	if (fdb_entry != NULL) 
        {
            if (fdb_entry->is_local == 1)
            {
                /* Hit. Local entry so pass the packet to the IP stack. */
	            br_pass_frame_up(br, skb);
            }
            else
            {
                /* Hit, but non local so push the packet on the appropriate interface. */
        		br_send_frame(fdb_entry->dst, skb);
            }

            br_fdb_put(fdb_entry);
    		return;
	    }

        /* No hit. Flood packet on all interfaces. Do not send the packet to the IP stack. */
	    br_flood(br, skb, 0);
    }
    else
    {
        /* Port was not in the forwarding state. So drop the packet and clean memory. */
		br_input_error(skb);
    }
    return;
}

/**************************************************************************
 * FUNCTION NAME : br_handle_frame
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered handler to which all packets shall be 
 *  pushed to. It locks the bridge and calls the actual bridge handler 
 *  routine.
***************************************************************************/
void br_handle_frame(struct sk_buff *skb)
{
	struct net_bridge *br;

    /* Trace Point: Start */ 
    psp_trace (BR_HANDLE_FRAME_ENTER);

    /* Get the bridge handle. */
	br = skb->dev->br_port->br;
	read_lock(&br->lock);

    /* Process the received frame. */
	__br_handle_frame(skb, br);

    /* Unlock the bridge. */
	read_unlock(&br->lock);

    /* Trace Point: Stop */
    psp_trace (BR_HANDLE_FRAME_EXIT);
    return;
}

/**************************************************************************
 * FUNCTION NAME : br_input_error
 **************************************************************************
 * DESCRIPTION   :
 *  The function is called on the bridge receive path when an error is 
 *  detected. It cleans up packet memory.
 ***************************************************************************/
static void br_input_error(struct sk_buff *skb)
{
    kfree_skb(skb);
    return;
}
