/* $Id: sock.c,v 1.39 2002/10/03 20:50:59 rjkaes Exp $
 *
 * Sockets are created and destroyed here. When a new connection comes in from
 * a client, we need to copy the socket and the create a second socket to the
 * remote server the client is trying to connect to. Also, the listening
 * socket is created and destroyed here. Sounds more impressive than it
 * actually is.
 *
 * Copyright (C) 1998  Steven Young
 * Copyright (C) 1999  Robert James Kaes (rjkaes@flarenet.com)
 * Copyright (C) 2000  Chris Lightfoot (chris@ex-parrot.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "tinyproxy.h"

#include "log.h"
#include "heap.h"
#include "sock.h"
#include "text.h"

#define DNS_LOCAL_QUERY
#ifdef DNS_LOCAL_QUERY
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <sys/select.h>

#ifdef SOCK_DEBUG
#define dbprint(x...) log_message(LOG_ERR, x)
#else
#define dbprint(x...)
#endif
#define DNS_BUFSIZE 512
#define DNS_LOCAL_TIMEOUT {0,100000}
struct dns_header 
{
	char id[2];
	char flags[2];
	char questions[2];
	char answers[2];
	char authority[2];
	char additional[2];
	char data[0];
};
#ifdef SOCK_DEBUG
char *dns_rcodes [] = 
{
	"No error", "Format error", "Server failure", "Name Error", "Not Implemented", "Refused"
};
#endif


char dnsLocal_buf[DNS_BUFSIZE];
int get_rcode(struct dns_header *hdr)
{
	return (hdr->flags[1] & 0xf);
}

// www.yahoo.com.sg         becomes 
// 3www5yahoo3com2sg0
int encode_domain(const char *in, char *out)
{
	int i, j;
	char *p, *q;
	int len;
 
 	p = (char *)in; q = out;
 	// skip any leading dots in input
	while(*p == '.') p++;
	
	len = strlen(p);
	
	for (i=0; i<len; i++)
	{
		for (j=0; p[i]!='.';i++,j++)
		{
			if (p[i] == 0) break;
			q[j+1] = p[i];
		}
		*q = j;
		q += (j+1);
	}
	*q = 0;
	return len+2;
}
int seq = 0;
int set_id(struct dns_header *hdr)
{
	short id = (getpid() & 0xffff) + (seq++);
	hdr->id[0] = (id >> 8) & 0xff;
	hdr->id[1] = id * 0xff;
	return (int) id;
}
#ifdef SOCK_DEBUG
int get_id(struct dns_header *hdr)
{
	return ((hdr->id[0] & 0xff)<<8 | (hdr->id[1] & 0xff)) & 0xffff;
}
#endif
int dnsLocal_mkquery (int op, const char *dname, int class,
              int type, char *data, int datalen, char *newrr,
              char *buf, int buflen)
{
	int qlen;
	struct dns_header *hdr = (struct dns_header *) buf;
	
	memset(hdr, 0, sizeof(struct dns_header));
	
	set_id(hdr);	

	// recursion desired
	hdr->flags[0] |= 1;
	hdr->questions[1] = 1;
	
	qlen = encode_domain(dname, hdr->data);
	hdr->data[qlen] = 0; hdr->data[qlen+1] = 1;
	hdr->data[qlen+2] = 0; hdr->data[qlen+3] = 1;
	
	return sizeof(struct dns_header) + qlen + 4;
}

int dnsLocal_sendrecv(char *server, char *buf, int len)
{
	int retval; int socklen;
	struct sockaddr_in sa;
	struct in_addr in;

	fd_set readfds; struct timeval tv = DNS_LOCAL_TIMEOUT;

	int fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	if (fd==-1)
	{
		dbprint("%s: Unable to open socket.\n", __FUNCTION__);
		return -1;	
	}
	if (!inet_aton(server, &in))
	{
		dbprint("%s: Bad IP server address %s\n", server);
		return -2;	
	}

	memset((void *)&sa, 0, sizeof(sa));
	memcpy(&sa.sin_addr.s_addr, &in, sizeof(in));
	sa.sin_port = htons(53);
	sa.sin_family = AF_INET;
	
	retval = sendto(fd, buf, len, 0, (struct sockaddr *)&sa, sizeof(sa));
	
	dbprint("%s: sent %d\n", __FUNCTION__, retval);

	/* still must select!! */
	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);
	
	retval = select(fd + 1, &readfds, NULL, NULL, &tv );
	if (retval<=0)
	{
		dbprint("%s: no response from server %s\n", __FUNCTION__, server);
		return -3;
	}
	dbprint("Time left: %d sec %d usec\n", tv.tv_sec, tv.tv_usec);
	
	socklen = sizeof(in);
	retval = recvfrom(fd, buf, DNS_BUFSIZE, 0, &sa, &socklen);
	dbprint("%s: DNS Response id=0x%04x length=%d\n", __FUNCTION__, get_id(buf), retval);
	close(fd);

	return retval;
}
int dnsLocal_query(const char *name, char *server, struct in_addr *addr)
{
	int len; int replylen; int rcode;
	char *p = dnsLocal_buf;
	char newrr[1]; char data[1];

	len = dnsLocal_mkquery(/*op*/ 0, /*dname*/ name, /*class*/ 1, 
	 /*type*/ 1, /*data*/ data, /*datalen*/ 1, /*newrr*/ newrr, 
	 /*buf*/ p, /*buflen*/ 512);
	dbprint("%s: DNS Query id=0x%04x length=%d\n", __FUNCTION__, get_id(p), len);

	replylen = dnsLocal_sendrecv(server, p, len);
	if (replylen<=0)
	{
		dbprint("%s: dnsLocal_sendrecv() failed\n", __FUNCTION__);
		return -1;	
	}

	rcode = get_rcode((struct dns_header *)p);
	if (rcode!=0)
	{
		dbprint("%s: Error %d (%s) in response\n", __FUNCTION__, rcode, dns_rcodes[rcode]);
		return -1;	
	}

	memcpy(addr, &p[replylen-sizeof(struct in_addr)], sizeof(struct in_addr));
	dbprint("%s: %s=%s\n", __FUNCTION__, name, inet_ntoa(*addr));

	return 0;
}

#endif

#ifdef NAMECACHE_ENABLE
/**
	khcheng - cache dns lookups
**/
#define NAMECACHE_SIZE 32
struct dns_cache {
	char *domain;
	struct in_addr addr;
};

static struct dns_cache nameCache[NAMECACHE_SIZE];
static int nameCache_index;
static int nameCache_initdone = 0;
static int nameCache_init(void)
{
	if (nameCache_initdone==1)
		return 0;

	log_message(LOG_ERR, "%s",__FUNCTION__);
	
	nameCache_index = 0;
	memset(nameCache, 0, sizeof(nameCache));
	
	nameCache_initdone = 1;
	return 0;
}
static int nameCache_add(char *name, struct in_addr *addr)
{
	char *p;
	struct in_addr *a;
	
	if (!name || !addr)
		return -1;

	log_message(LOG_ERR, "%s: %s=%s",__FUNCTION__, name, inet_ntoa(*addr));

	p = nameCache[nameCache_index].domain;
	if (p)
	{
		free(p);
	}
	p = strdup(name);

	a = &(nameCache[nameCache_index].addr);
	memcpy(a, addr, sizeof(struct in_addr));
		
	nameCache_index = ((nameCache_index + 1) % NAMECACHE_SIZE);
	return 0;
}
static int nameCache_find(char *name, struct in_addr *addr)
{
	int i;
	char *p;
	log_message(LOG_ERR, "%s: %s",__FUNCTION__, name);
	for (i=0; i<NAMECACHE_SIZE; i++)
	{
		if ((p = nameCache[i].domain))
		{
			log_message(LOG_ERR, "%s: #%d %s=%s",__FUNCTION__, i, p, inet_ntoa(nameCache[i].addr));
			if (!strcmp(name, p))
			{
				memcpy(addr, &(nameCache[i].addr), sizeof(struct in_addr));
				log_message(LOG_ERR, "%s: %s=%s",__FUNCTION__, name, inet_ntoa(nameCache[i].addr));
				return 0;
			}
		}
	}
	return -1;
}
#endif
/*
 * Take a string host address and return a struct in_addr so we can connect
 * to the remote host.
 *
 * Return a negative if there is a problem.
 */
static int
lookup_domain(struct in_addr *addr, const char *domain)
{	
	struct hostent* result;

	assert(domain != NULL);

	#ifdef NAMECACHE_ENABLE
	nameCache_init();
	if (0==nameCache_find(domain, addr))
	{
		return 0;
	}
	#endif

	#ifdef DNS_LOCAL_QUERY
	if(0==dnsLocal_query(domain, "127.0.0.1", addr))
	{
		return 0;
	}
	#endif
	
	result = gethostbyname(domain);
	if (result) {
		memcpy(addr, result->h_addr_list[0], result->h_length);
		#ifdef NAMECACHE_ENABLE
		nameCache_add(domain, addr);
		#endif
		return 0;
	} else
		return -1;
}

/* This routine is so old I can't even remember writing it.  But I do
 * remember that it was an .h file because I didn't know putting code in a
 * header was bad magic yet.  anyway, this routine opens a connection to a
 * system and returns the fd.
 *	- steve
 *
 * Cleaned up some of the code to use memory routines which are now the
 * default. Also, the routine first checks to see if the address is in
 * dotted-decimal form before it does a name lookup.
 *      - rjkaes
 */
int
opensock(char *ip_addr, uint16_t port)
{
	int sock_fd;
	struct sockaddr_in port_info;
	struct sockaddr_in bind_addr;
	int ret  ;

	assert(ip_addr != NULL);
	assert(port > 0);

	memset((struct sockaddr *) &port_info, 0, sizeof(port_info));

	port_info.sin_family = AF_INET;

	/* Lookup and return the address if possible */
	ret = lookup_domain(&port_info.sin_addr, ip_addr);
	if (ret < 0) {
		log_message(LOG_ERR,
			    "opensock: Could not lookup address \"%s\".",
			    ip_addr);
		return -1;
	}

	port_info.sin_port = htons(port);

	if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		log_message(LOG_ERR, "opensock: socket() error \"%s\".",
			    strerror(errno));
		return -1;
	}

	/* Bind to the specified address */
	if (config.bind_address) {
		memset(&bind_addr, 0, sizeof(bind_addr));
		bind_addr.sin_family = AF_INET;
		bind_addr.sin_addr.s_addr = inet_addr(config.bind_address);

		ret = bind(sock_fd, (struct sockaddr *)&bind_addr, sizeof(bind_addr));
		if (ret < 0) {
			log_message(LOG_ERR, "Could not bind local address \"%\" because of %s",
				    config.bind_address,
				    strerror(errno));
			return -1;
		}
	}

	if (connect(sock_fd, (struct sockaddr *) &port_info, sizeof(port_info)) < 0) {
		log_message(LOG_ERR, "opensock: connect() error \"%s\".",
			    strerror(errno));
		return -1;
	}

	return sock_fd;
}

/*
 * Set the socket to non blocking -rjkaes
 */
int
socket_nonblocking(int sock)
{
	int flags;

	assert(sock >= 0);

	flags = fcntl(sock, F_GETFL, 0);
	return fcntl(sock, F_SETFL, flags | O_NONBLOCK);
}

/*
 * Set the socket to blocking -rjkaes
 */
int
socket_blocking(int sock)
{
	int flags;

	assert(sock >= 0);

	flags = fcntl(sock, F_GETFL, 0);
	return fcntl(sock, F_SETFL, flags & ~O_NONBLOCK);
}

/*
 * Start listening to a socket. Create a socket with the selected port.
 * The size of the socket address will be returned to the caller through
 * the pointer, while the socket is returned as a default return.
 *	- rjkaes
 */
int
listen_sock(uint16_t port, socklen_t* addrlen)
{
	int listenfd;
	const int on = 1;
	struct sockaddr_in addr;

	assert(port > 0);
	assert(addrlen != NULL);

	listenfd = socket(AF_INET, SOCK_STREAM, 0);
	setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);

	if (config.ipAddr) {
		addr.sin_addr.s_addr = inet_addr(config.ipAddr);
	} else {
		addr.sin_addr.s_addr = inet_addr("0.0.0.0");
	}

	if (bind(listenfd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		log_message(LOG_ERR, "Unable to bind listening socket because of %s",
			    strerror(errno));
		return -1;
	}

	if (listen(listenfd, MAXLISTEN) < 0) {
		log_message(LOG_ERR, "Unable to start listening socket because of %s",
			    strerror(errno));
		return -1;
	}

	*addrlen = sizeof(addr);

	return listenfd;
}

/*
 * Return the peer's socket information.
 */
int
getpeer_information(int fd, char* ipaddr, char* string_addr)
{
	struct sockaddr_in name;
	size_t namelen = sizeof(name);
	struct hostent* result;

	assert(fd >= 0);
	assert(ipaddr != NULL);
	assert(string_addr != NULL);

	/*
	 * Clear the memory.
	 */
	memset(ipaddr, '\0', PEER_IP_LENGTH);
	memset(string_addr, '\0', PEER_STRING_LENGTH);

	if (getpeername(fd, (struct sockaddr *)&name, &namelen) != 0) {
		log_message(LOG_ERR, "getpeer_information: getpeername() error: %s",
			    strerror(errno));
		return -1;
	} else {
		strlcpy(ipaddr,
			inet_ntoa(*(struct in_addr *)&name.sin_addr.s_addr),
			PEER_IP_LENGTH);
	}
	#ifndef DNS_REVERSE_LOOKUP
	return 0;
	#else
	result = gethostbyaddr((char *)&name.sin_addr.s_addr, 4, AF_INET);
	if (result) {
		strlcpy(string_addr, result->h_name, PEER_STRING_LENGTH);
		return 0;
	} else {
		return -1;
	}
	#endif
}
